#include <stdio.h>
#include <stdlib.h>
#include <FL/fl_utf8.h>
//#include "flstring.h"

#include <FL/Fl.H>
#include "../FL/Fl_Shareable_Image.H"
#include <FL/Fl_XBM_Image.H>
#include <FL/Fl_XPM_Image.H>
#include <FL/fl_draw.H>
#include <math.h>


    // COLOR EFFECT
    //
    // Sequential effects can be easily combined to single "effect matrix" because both
    // averaging and desaturation are sort of linear color transformation.

    // COLOR AVERAGING
    // ---------------
    //
    // For color averaging we have r = F * c + (1-F) * CA
    // where c is original color, F is multiplication factor, CA averaging color.
    // We can redefine above  to   r = (1-f) * c  + ca  where
    //     f = 1-F   (F = 1-f)
    //     ca = (1-F) * CA = f * CA
    // In the "effect matrix" we store f and ca (instead of F and CA).
    //
    // Averaging can be written as multiplication by a matrix M
    //   R = C * M ;    R(r) = C(c) * M(F, ca)
    //                           | F    0 |
    //   | r   1 | = | c   1 | * |        |
    //                           | ca   1 |
    //
    // Ssquential effect can be then combined as matrix multiplication, for two we get
    //    R = C * M1(F1, ca1) * M2 (F2, ca2)
    //
    //                         | F1   0 |   | F2   0 |
    //   | r  1 | = | c  1 | * |        | * |        |
    //                         | ca1  1 |   | ca2  1 |
    //
    //  and combined effect matrix is
    //                                          | F1*F2       0 |
    //   M(F, ca) = M1(F1, ca1) * M2(F2, ca2) = |               |
    //                                          | F2*ca1+ca2  1 |
    //
    // In "stored" parameters that means
    //   f = 1 - (1 - f1) * (1 - f2)
    //   ca = (1 - f2) * ca1 + ca2
    //
    //
    // DESATURATION
    // ------------
    //
    // We denode desaturation operator d on color c
    //    r =  c @ d       where for d == 0    c @ d = c
    //                           for d == 1    c @ d = desaturate(c) = (31*cr + 61*cg +8*cb)/100
    //
    // As matrix this desaturation can be written
    //   R = C * D(d)
    //                           | @d   0 |
    //   | r   1 | = | c   1 | * |        |
    //                           |  0   1 |
    //
    // Desaturation is associative with respect to averaging matrix, that means
    //    R = C * M(F, ca) @ D(s) = (C @ D(s))  *  (M(F, ca) @ D(s)))  or
    //    R  = (C @ d(s)) * M(F, ca @ d(s))
    // This means that  desaturation propagates to the front and we need to
    // recalculate/desaturate only averaging color.
    //
    // In other words :
    // Averaging and then desaturation can be swapped to desaturation first and then
    // averaging with desaturated color. That also means that any sequence of averaging
    // and desaturation can be combined to "desaturation first" and then averaging with
    // combined/recalculated averaging color.
    // We generalise both desaturation and averaging matrices as
    //
    //      | @d*F  0 |
    // M' = |         |
    //      | ca    1 |
    //
    // and because multiplying r,g,b channels by the same factor F (not matrix)
    // and desaturating is the same as desaturating first and then multiplying
    // and because @d1 @d2 = @(d1|d2) (we desaturate image if either matrix is desaturated)
    // final form of matrix multiplication and combined effect matrix is
    //
    //                                                        | @(d1|d2)*F1*F2  0 |
    //   M'(d, F, ca) = M1'(d1, F1, ca1) * M2'(d2, F2, ca2) = |                   |
    //                                                        | ca1@d2*F2+ca2   1 |
    //
    // So for each multiplication step
    // we calculate combination as
    //    d = d1 | d2
    //    F = F1 * F2
    //    ca = ca1@d2*F2 + ca2      where   x@d = x  ....       for d = 0
    //                                      x@d = desaturate(x) for d = 1


// Some day following should go to fl_draw.h
// The color effect pseudo-matrix is encoded such as:
// - The most significant bit represents possible desaturation
// - Next seven bits in most significant BYTE is averaging factor,
//   clamped and rounded to 127 levels as X = (1-F) * 127
// - The remaining three bytes is averaging color.
// This means that "no effect" (identity matrix) has value zero
// which is convenient for various conditional testing:
//  effect == 0 ... no operation required
//  (effect & FL_COLOR_EFFECT_DESATURATE) == 0 ... no desaturation required
//  (effect & FL_COLOR_EFFECT_AVERAGE) == 0 ... no averaging required


enum Fl_Color_Effect {
    FL_COLOR_EFFECT_NONE = 0,
    FL_COLOR_EFFECT_DESATURATE = 0x80000000,
    FL_COLOR_EFFECT_AVERAGE = 0x7F000000
};

// Combines effect \a e1 with following effect \a e2 and returns combined effect
Fl_Color_Effect fl_color_effect_multiply(Fl_Color_Effect e1, Fl_Color_Effect e2);

// Appends desaturation to the effect (possibly desaturating averaging color)
// and returns combined effect. If e = 0, it creates the effect
// (the same as combining with previous FL_COLOR_EFFECT_NONE)
Fl_Color_Effect fl_color_effect_desaturate(Fl_Color_Effect e = FL_COLOR_EFFECT_NONE);

// Creates color-averaging effect
Fl_Color_Effect fl_color_effect_average(Fl_Color c, float mix);

// Adds color-averaging to the effect e and returns combined effect
inline Fl_Color_Effect fl_color_effect_average(Fl_Color_Effect e, Fl_Color c, float mix){
    return fl_color_effect_multiply(e, fl_color_effect_average(c, mix));
}

// Retreieves required operations from the effect \a e.
// If \a desaturate is true, desaturation should be performed first following by
// color-averaging with \a average_color and \a average_factor F.
// Note: you can check what operations are required by effect masking:
// - if (e & FL_COLOR_EFFECT_DESATURATE) desaturation is required
//   (this is what is returned in \a desaturate)
// - if (e & FL_COLOR_EFFECT_AVERAGE) color-averaging is required
//   (this is simpler than checking if returned \a average_factor is not 1)
void fl_color_effect_get(Fl_Color_Effect e, int &desaturate, Fl_Color &average_color, float &average_factor);

// Applies color effect \a e on color \a c and returns resulting color
Fl_Color fl_color_effect(Fl_Color_Effect e, Fl_Color c);




void fl_color_effect_get(Fl_Color_Effect m_, int &desaturation, Fl_Color & average_color, float & factor) {
  int f = m_ & FL_COLOR_EFFECT_AVERAGE;
  desaturation = (m_ & FL_COLOR_EFFECT_DESATURATE);
  if(!f) {
    average_color = Fl_Color(0);
    factor = 1.0;
    return;
  }
  f >>= 24;
  factor = (float(127 - f)) / 127.0f;
  int c = (((m_ & 0xFF0000)>>16)*127)/f;
  int ret = c<<24;
  c = (((m_ & 0xFF00)>>8)*127)/f;
  ret |= c<<16;
  average_color = Fl_Color(ret | ((((m_ & 0xFF)*127)/f)<<8));
}


Fl_Color_Effect fl_color_effect_desaturate(Fl_Color_Effect m) {
  // if the image was averaged we need also desaturate this averaging addition too
  if(!(m & FL_COLOR_EFFECT_AVERAGE))
    return FL_COLOR_EFFECT_DESATURATE;
  int des_c = (31 * ((m & 0xFF0000)>>16) + 61 * ((m & 0xFF00)>>8) + 8 * (m & 0xFF) + 50)/100;
  return Fl_Color_Effect(des_c | (des_c <<8) | (des_c << 16) | (m &  0xFF000000) | FL_COLOR_EFFECT_DESATURATE);
}


Fl_Color_Effect fl_color_effect_multiply(Fl_Color_Effect m1, Fl_Color_Effect m2) {
  if(m2 & FL_COLOR_EFFECT_DESATURATE)
    m1 = fl_color_effect_desaturate(m1);

  int f2 = m2 & FL_COLOR_EFFECT_AVERAGE;
  if(!f2)
    return m1;
  int f1 = m1 & FL_COLOR_EFFECT_AVERAGE;
  if(!f1)
    return Fl_Color_Effect((m1 & FL_COLOR_EFFECT_DESATURATE) | (m2 & 0x7FFFFFFF));
  f1 >>= 24;
  f2 >>= 24;
  int F1 = 127 - f1;
  int F2 = 127 - f2;
  int f = 127 - ((F1 * F2+63)/127);
  int ret = (m1 & FL_COLOR_EFFECT_DESATURATE) | (f<<24);
  // ca = F2 * ca1 + ca2
  int ca1 = (m1 & 0xFF0000)>>16;
  int ca2 = (m2 & 0xFF0000)>>16;
  int ca = ((F2 * ca1 + 63)/127) + ca2;
  ret |= ca <<16;
  ca1 = (m1 & 0xFF00)>>8;
  ca2 = (m2 & 0xFF00)>>8;
  ca = ((F2 * ca1 + 63)/127) + ca2;
  ret |= ca << 8;
  ca1 = m1 & 0xFF;
  ca2 = m2 & 0xFF;
  ca = ((F2 * ca1 + 63)/127) + ca2;
  return Fl_Color_Effect(ret | ca);
}


Fl_Color_Effect fl_color_effect_average( Fl_Color c, float factor) {
  int F = int(127.0f * factor + 0.5);
  if(F>=127)
    return FL_COLOR_EFFECT_NONE;
  if(F<0) F = 0;
  int f = 127 - F;
  int CA = Fl::get_color((Fl_Color)c)>>8;
  int ca = ((((CA>>16) & 0xFF) * f + 63)/127) <<16;
  ca |= ((((CA>>8) & 0xFF) * f + 63)/127) <<8;
  return Fl_Color_Effect((f<<24) | ca | (((CA & 0xFF) * f + 63)/127));
}


Fl_Color fl_color_effect(Fl_Color_Effect m_, Fl_Color i) {
  int c = Fl::get_color(i);
  int r = (c & 0xFF000000)>>24;
  int g = (c & 0xFF0000)>>16;
  int b = (c & 0xFF00)>>8;
  int f = m_ & FL_COLOR_EFFECT_AVERAGE;
  int F = 127 - (f>>24);
  if(m_ & FL_COLOR_EFFECT_DESATURATE) {
    int a = (31 * r + 61 * g + 8 * b);
    if(f) {
      a *= F;
      int k = 100 * 127;
      r = (a +  k * (( m_ & 0xFF0000)>>16))/k;
      g = (a +  k * (( m_ & 0xFF00)>>8))/k;
      b = (a +  k *( m_ & 0xFF ))/k;
      return Fl_Color((r<<24) | (g<<16) | (b<<8));
    } else {
      a /= 100;
      return Fl_Color((a<<24) | (a<<16) | (a<<8));
    }
  }
  if(!f) return i;
  r *= F; r += 63;
  r /= 127;
  r += (m_ & 0xFF0000)>>16;
  g *= F; g += 63;
  g /=127;
  g += (m_ & 0xFF00)>>8;
  b *= F; b +=63;
  b /=127;
  b += (m_ & 0xFF);
  return Fl_Color((r<<24) | (g<<16) | (b<<8));
}


//static int assure_data_ = 0;











#define RI(i)((Fl_Registered_Image *)(i))
#define CRI(i)((const Fl_Registered_Image *)(i))


void fl_display_shared();  // for testing purposes



// This is necessary because we can not reuse fl_check_images in fl_images_core (it is static there)
// When Fl_Shared_Images will be replaced by above implementation, this can be removed.
#include <FL/Fl_BMP_Image.H>
#include <FL/Fl_GIF_Image.H>
#include <FL/Fl_JPEG_Image.H>
#include <FL/Fl_PNG_Image.H>
#include <FL/Fl_PNM_Image.H>

// temporary hack
#ifndef HAVE_LIBPNG
#define HAVE_LIBPNG
#endif
#ifndef HAVE_LIBJPEG
#define HAVE_LIBJPEG
#endif

static Fl_Image *					// O - Image, if found
fl_check_images_(const char *name,		// I - Filename
                uchar      *header,		// I - Header data from file
		int) {				// I - Amount of data (not used)
  if (memcmp(header, "GIF87a", 6) == 0 ||
      memcmp(header, "GIF89a", 6) == 0)	// GIF file
    return new Fl_GIF_Image(name);

  if (memcmp(header, "BM", 2) == 0)	// BMP file
    return new Fl_BMP_Image(name);

  if (header[0] == 'P' && header[1] >= '1' && header[1] <= '7')
					// Portable anymap
    return new Fl_PNM_Image(name);

#ifdef HAVE_LIBPNG
  if (memcmp(header, "\211PNG", 4) == 0)// PNG file
    return new Fl_PNG_Image(name);
#endif // HAVE_LIBPNG

#ifdef HAVE_LIBJPEG
  if (memcmp(header, "\377\330\377", 3) == 0 &&
					// Start-of-Image
      header[3] >= 0xc0 && header[3] <= 0xef)
	   				// APPn for JPEG file
    return new Fl_JPEG_Image(name);
#endif // HAVE_LIBJPEG
  return 0;
}


#define DS fl_display_shared();



namespace { // anonymous internal class for registered images




// This is purely internal class, however it is a subclass of Fl_Shareable_Image   for backeard comparibility
class Fl_Registered_Image: public Fl_Shareable_Image {
  friend class Fl_Shareable_Image;
  friend void fl_display_shared_();
  static Fl_Registered_Image ** images_;
  static int num_images_;
  static int alloc_images_;

  static int compare(Fl_Registered_Image **i0, Fl_Registered_Image **i1);
  static int bsearch_position (Fl_Registered_Image * *key, Fl_Registered_Image * * base, int num,   int (*compar)(Fl_Registered_Image **,Fl_Registered_Image **), int &position);

  const char * name_;			// Name of image file
  int original_;
  int refcount_;
  int alloc_image_;

  void add_(int index);
  int add (Fl_Shareable_Image *); // for backward compatibility only
  void release();
  void reload();
  void reference() {refcount_++;}
  void dereference() {refcount_--;}
  // We never uncache in case data are not assured
  // void uncache(){ if (image_ && ((MUST_HAVE_DATA & flags_) || original_)) image_->uncache();}
  ~Fl_Registered_Image();
  Fl_Registered_Image * lazy_copy(int W, int H, Fl_Color_Effect lazy_effect_, const char * new_name = 0);
  static int find_(const char *n, Fl_Color_Effect effect, int W, int H, int &position, int & original_position);
  Fl_Registered_Image(const char *n, Fl_Image   *img = 0);
  Fl_Registered_Image();
   // after lazy_copy() this assures sreation of internal image
  void apply_effect(Fl_Color_Effect e); // low level application of effect on image but without any other member change
  int flags_;  // for the moment osed only to allow lazy copy

  // New fields
  Fl_Registered_Image * original_image_;
  Fl_Color_Effect effect_;
  Fl_Color_Effect lazy_effect_; // to be applied after copying from original real image is created

  Fl_Shareable_Image * to_remove; // only for backward-compatibility when Fl_Shareable_Image uses add()

  void update_data() {
  if (image_) {
    w(image_->w());
    h(image_->h());
    d(image_->d());
    Fl_Image::data(image_->data(), image_->count());
  }
}



public:
  Fl_Image * internal_image() const {return image_;}
  void set_sizes(int W, int H){w(W); h(H);} // need public for fl_draw(...)
  void update_image();
  void update_shared(Fl_Shareable_Image *);
  Fl_Color_Effect effect() const {return effect_;}
  static Fl_Registered_Image * find(const char *n, Fl_Color_Effect effect, int W, int H);
  static Fl_Registered_Image * get(const char *n, Fl_Color_Effect effect, int W, int H);
  Fl_Registered_Image * get(int W, int H, Fl_Color_Effect additional_effect);
  void draw(int X, int Y, int W, int H, int cx, int cy);
  bool is_bitmap();

  enum Flags{
     MUST_HAVE_DATA = 1
  };
  void set_flags(int f){ flags_ |=f;}
  void clear_flags(int f) {flags_ &= ~f;}

  void assure_data(){
      set_flags(MUST_HAVE_DATA);
      update_image();
  }

};


Fl_Registered_Image **Fl_Registered_Image::images_ = 0;	// Shared images
int	Fl_Registered_Image::num_images_ = 0;	// Number of shared images
int	Fl_Registered_Image::alloc_images_ = 0;	// Allocated shared images

static Fl_Registered_Image * fl_original_image_ = 0;
static const Fl_Image * fl_original_for_ = 0;
static void set_fast_original(const Fl_Image * for_image, Fl_Registered_Image * original){fl_original_for_ = for_image, fl_original_image_ = original;}

static Fl_Image * (*copy_function_)(Fl_Image * im, int W, int H) = 0;


Fl_Image * copy_image(Fl_Image * im, int W, int H){
  if(!copy_function_) return im->copy(W, H);
  Fl_Image * i = (*copy_function_)(im, W, H);
  if(i) return i;
  return im->copy(W, H);
}


bool Fl_Registered_Image::is_bitmap(){
  Fl_Registered_Image * im = this;
  while(!im->image_ && im->original_image_!=im)
    im = im->original_image_;
  if(im->image_ && !im->image_->d()) return true;
  return false;
}


int Fl_Registered_Image::compare(Fl_Registered_Image **i0, Fl_Registered_Image **i1) {
    int i = strcmp((*i0)->name_, (*i1)->name_);
    if (i) return i;
    i = (*i0)->effect_ - (*i1)->effect_;
    if(i) return i;
    if (((*i0)->w() == 0 && (*i1)->original_))
      return 0;
    i = (*i0)->w() - (*i1)->w();
    if(i) return ((*i1)->original_)? ((*i1)->original_) : i;
    return (*i0)->h() - (*i1)->h();

  }

// This is my version of "bsearch" which - unlike standard bsearch - also returns "insertion position" if the item is not found.
// That means that expensive qsort is not required because if an image is not found, newly created image is inserted at the position
// returned by previous bsearch call.
// If found, returns 0 and the place is returned by \a position argument. If not found it returns non-zero and gives "insertion"
// place in position. If new element should be added with corresponding key, everything from that position
// (including) should be shifted up and the element should be inserted there without a need of further sorting.
// If you wish to insert "after" the element in a case of match, you can increment position depending on return value:
// if(ret_value>=0) position++;

int Fl_Registered_Image::bsearch_position (Fl_Registered_Image * *key, Fl_Registered_Image * * base, int num,   int (*compar)(Fl_Registered_Image **,Fl_Registered_Image **), int &position){
  if(!num){
    position = 0;
    return -1;
  }
  int s = 0;
  int e = num - 1;
  int add =(e-s)>>1;
  while(add){
    int m = s + add;
    int c = (*compar)(key, base +  m);
    if(c<0)
      e = m;
    else if (c>0)
      s = m;
    else{
      position = m;
      return 0;
    }
    add = (e-s)>>1;
  }
  position =  s;
  if(!s){
    int ret = (*compar)(key, base);
    if(ret<=0) return ret;
  }
  if(e && e==(num-1)){
    int ret = (*compar)(key, base + e);
    if(ret>=0){
      position = e;
      return ret;
    }
  }
  return 1;
}


Fl_Registered_Image::Fl_Registered_Image() : Fl_Shareable_Image() {
  name_        = 0;
  refcount_    = 1;
  original_    = 0;
  image_       = 0;
  alloc_image_ = 0;
  original_image_ = 0;
  to_remove = 0;
  // if(assure_data_)
  //  flags_ = MUST_HAVE_DATA ;
  // else
    flags_ = 0;
  effect_ = FL_COLOR_EFFECT_NONE;
  lazy_effect_ = FL_COLOR_EFFECT_NONE;
}


// O
Fl_Registered_Image::Fl_Registered_Image(const char *n,		// I - Filename
                      Fl_Image   *img)	// I - Image
    : Fl_Shareable_Image() {
    name_ = new char[strlen(n) + 1];
    strcpy((char *)name_, n);

    refcount_    = 1;
    image_       = img;
    alloc_image_ = !img;
    original_    = 1;
    to_remove = 0;
    original_image_ = this;
    effect_ = FL_COLOR_EFFECT_NONE;
    lazy_effect_ = FL_COLOR_EFFECT_NONE;

//    if(assure_data_)
//      flags_ = 1;
//    else
      flags_ = 0;
    original_image_ = this;

    if (!img)
      reload();
    else
      update_data();
  }


int Fl_Registered_Image::add(Fl_Shareable_Image * s){
   if(to_remove) return 1; // already here
   to_remove = s;
   int position, original_position;
   int found = find_(name_, effect_ , w(), h(), position, original_position);
   if(found)
      return 2; // sorry, can't mess with already inserted images with this name
   add_(position);
   return 0;
}

void Fl_Registered_Image::add_( int index) {
  Fl_Registered_Image	**temp;		// New image pointer array...
  const int extra_space = 64;


  if (num_images_ >= alloc_images_) {
    // Allocate more memory...
    int to_alloc = (alloc_images_ * 3) / 2 + extra_space;
    temp = new Fl_Registered_Image *[to_alloc];
    if (alloc_images_) {
      memcpy(temp, images_, alloc_images_ * sizeof(Fl_Registered_Image *));
      delete[] (Fl_Registered_Image * *)images_;
    }
    images_       = temp;
    alloc_images_ = to_alloc;
  }

  if(index<num_images_){
    memmove(images_ + index + 1, images_ + index, (num_images_ - index) * sizeof(Fl_Registered_Image *));
    images_[index] = this;
    num_images_ ++;
  }else{
    images_[index] = this;
    num_images_ ++;
  }

}


Fl_Registered_Image::~Fl_Registered_Image() {
  if (original_ && name_) delete[] (char *)name_;
  if(image_ && alloc_image_)
      delete image_;
  if(original_image_ && original_image_!=this) // releasing original
    original_image_->release();
  if(to_remove && ((Fl_Registered_Image *)to_remove)->image_==this){ // this is only baskward-compatibility hack
    RI(to_remove)->name_ = 0; // we must assure than name is not destroyed for shared image because we are changing image_
    RI(to_remove)->image_ = 0;
    delete to_remove;
  }
  // we need to do some cleanup so ~Fl_Shareable_Image()
  // destructor does not mess with us trying to release...
  name_ = 0;
  image_ = 0;
}

void Fl_Registered_Image::release() {
  int	i;	int orig_pos;
  dereference();
  if (refcount_ > 0) return;
  // if(alloc_image_) { // RK: this was another bug which was wringly set if there was This is a fix
    if(0>= find_(name_, effect_, w(), h(), i, orig_pos) || (images_[i]!=this))
      for (i = 0; i < num_images_; i ++) // fallback, slower for lot of images .... could be removed
        if (images_[i] == this) break;
    if(i<num_images_){
      num_images_ --;
      if (i < num_images_) {
        memmove(images_ + i, images_ + i + 1,
        (num_images_ - i) * sizeof(Fl_Registered_Image *));
      }
    }
    if (num_images_ == 0 && images_) {
      delete[] images_;
      images_       = 0;
      alloc_images_ = 0;
    }
  //}
  delete this;
}

void Fl_Registered_Image::reload() {
  if(!original_) return;
  // Load image from disk...
  int		i;		// Looping var
  FILE		*fp;		// File pointer
  uchar		header[64];	// Buffer for auto-detecting files
  Fl_Image	*img;		// New image
  if(!alloc_image_) return;

  if (!name_) return;
 // const char * const * d = 0;
 // if(image_)
 //   d = image_->data();

  if ((fp = fl_fopen(name_, "rb")) != NULL) {
    if (fread(header, 1, sizeof(header), fp)==0) { }
    fclose(fp);
  } else {
    return;
  }

  // Load the image as appropriate...
  if (memcmp(header, "#define", 7) == 0) // XBM file
    img = new Fl_XBM_Image(name_);
  else if (memcmp(header, "", 9) == 0) // XPM file
    img = new Fl_XPM_Image(name_);
  else {
    // Not a standard format; try an image handler...
    for (i = 0, img = 0; i < num_handlers_; i ++) {
      img = (handlers_[i])(name_, header, sizeof(header));

      if (img) break;
    }
  }

  if (img) {
    if (alloc_image_) delete image_;

    alloc_image_ = 1;

    if ((img->w() != w() && w()) || (img->h() != h() && h())) {
      // Make sure the reloaded image is the same size as the existing one.
      Fl_Image *temp = copy_image(img, w(), h());
      //Fl_Image *temp = img->copy(w(), h());
      delete img;
      image_ = temp;
    } else {
      image_ = img;
    }

    update_data();
  }
}


Fl_Registered_Image* Fl_Registered_Image::find(const char *n, Fl_Color_Effect effect, int W, int H){
  int position, original;
  //printf("Finding: %s\n", n);
  int f = find_(n, effect, W, H, position, original);
  if(f>0) return (Fl_Registered_Image *)(images_[position]);
  return 0;
}



// Returns positive if found, negative if found only original, othervise 0
// If not found but original exists, returns index of the original.
// Knowing positions if search fails can faster because if we know where to insert new image(s) and we dont need to perform expensive qsort
int Fl_Registered_Image::find_(const char *n, Fl_Color_Effect effect, int W, int H, int &position, int & original_position) {
//  Fl_Registered_Image	*key;		// Image key
//			**match;	// Matching image
  static Fl_Registered_Image key_;
  original_position = 0;
  if (num_images_) {
    key_.name_ =  n;
    key_.w(W);
    key_.h(H);
    key_.effect_ = effect;
    Fl_Registered_Image * key_pt = &key_;
    int found = bsearch_position(&key_pt, (Fl_Registered_Image **) images_, num_images_, compare, position);


    if (!found) {
      key_.name_ =  0; // for clean exit so the destructor does not try to delete the name string
      ((Fl_Registered_Image *)(images_[position]))->reference();
      return 1;
    }
    if(found>=0)
      position++;

    if(W){ // it could happen than original was missed (which must have unique place to be searchable) and still has the same W and H
      key_.w(0);
      found = bsearch_position(&key_pt, (Fl_Registered_Image **)images_, num_images_, compare, original_position);
      if(!found){
        key_.name_ =  0;
        if(images_[original_position]->w() == W &&  images_[original_position]->h()==H){
          position = original_position;
          ((Fl_Registered_Image *)(images_[position]))->reference();
          return 1;
        }
        return -1;
      }
      if(found>0)
        original_position++;
    }
    key_.name_ =  0;

  }else{
    position = 0;
    original_position  = 0;
  }


  return 0;
}

void Fl_Registered_Image::update_image(){
    if(!image_){
        if(original_image_==this)
          return;
        // find an original with data
        Fl_Registered_Image * orig = original_image_;
        Fl_Color_Effect effect = lazy_effect_;
        while(!(orig->image_) && (orig->original_image_ != orig)){
          effect = fl_color_effect_multiply(orig->lazy_effect_, effect);
          orig = orig->original_image_;
        }
         image_ = copy_image(orig->image_, w(), h());
        //image_ = orig->image_->copy(w(), h()); // we are not poor anymore!
        if(effect)
          apply_effect(effect);          // need to do what we are told
        update_data();
    }
}

void Fl_Registered_Image::apply_effect(Fl_Color_Effect e){
    if(!image_ || !e) return;
    int desaturation;
    float factor;
    Fl_Color col;
    fl_color_effect_get(e, desaturation, col, factor);
    if(e & FL_COLOR_EFFECT_DESATURATE)
      image_->desaturate();
    if(e & FL_COLOR_EFFECT_AVERAGE)
      image_->color_average(col, factor);
}

Fl_Registered_Image * Fl_Registered_Image::get(const char *n, Fl_Color_Effect effect, int W, int H) {
  Fl_Registered_Image	*temp;		// Image
  int position, original_position;
  int found = find_(n, effect, W, H, position, original_position);
  if(found>0)
    return images_[position];
  if(found<0){ // found only original
    temp = images_[original_position];
    temp->reference();
  }else{
    if(effect){ // trying to find non-effected image
       int position_0, original_position_0;
       int found = find_(n, FL_COLOR_EFFECT_NONE , 0, 0, position_0, original_position_0);
       if(found>0) // ok, we have non-effected original
          temp = images_[position_0];
       else{
          temp = new Fl_Registered_Image(n); // loading from disk
          temp->add_(original_position_0);
          if(effect>0){ // we need to increment insertions because our effect will be after unaffected original
              position++;
              original_position++;
          }
       }
       temp = temp->lazy_copy(0, 0, effect); // for now temp depends on non-effected original
    }else
       temp = new Fl_Registered_Image(n);
    if (!temp->image_&& (!temp->original_image_ || temp->original_image_==temp)) {
       delete temp;
       return NULL;
    }
    temp->original_ = 1;
    temp->add_(original_position);
    if(!W || (temp->w()==W && temp->h()==H))
      return temp;
    position++; // we have just inserted original, copy must go further
  }
 // Fl_Registered_Image * from;
  temp = (Fl_Registered_Image *)temp->lazy_copy(W, H, FL_COLOR_EFFECT_NONE);
  temp->add_(position);
  return temp;
}



Fl_Registered_Image * Fl_Registered_Image::lazy_copy(int W, int H, Fl_Color_Effect lazy_effect, const char * new_name) {
  //Fl_Image		*temp_image;	// New image file
  Fl_Registered_Image	*temp_shared;	// New shared image
  if(!new_name) new_name = name_;
//  const char * const * d = 0;
  temp_shared = new Fl_Registered_Image();
  temp_shared->original_image_ = this;
  temp_shared->lazy_effect_ = lazy_effect;
  temp_shared->effect_ = fl_color_effect_multiply(effect_, lazy_effect);
  temp_shared->w(W? W : w());
  temp_shared->h(H? H : h());
  if(W)
    temp_shared->name_ = new_name;
  else{ // oh yeah, he might want ALL the pixels some day
    temp_shared->original_ = 1;
    temp_shared->name_ = new char[strlen(new_name) + 1];
    strcpy((char *)temp_shared->name_, new_name);
  }
 temp_shared->refcount_    = 1;
 temp_shared->alloc_image_ = 1;
 if(temp_shared->flags_ & 1)
    temp_shared->update_image(); // even worse, he is strong enought to demand them now!
                                 // Lets give it to him while we are strong with refcaunts and can't kill us...
 return temp_shared;
}

Fl_Registered_Image * Fl_Registered_Image::get(int W, int H, Fl_Color_Effect additional_effect){
    if(!original_image_) return 0;
    if(!additional_effect){
        if(!W){
          original_image_->reference();
          return original_image_;
        }
        if(W==w() && H==h()){
          reference();
          return this;
        }
    }
    Fl_Color_Effect ne = fl_color_effect_multiply(effect_, additional_effect);
    int position, original_position;
    int found = find_(name_, ne, W, H, position, original_position);
    if(found>0)
      return images_[position];
    if(found<0){
      Fl_Registered_Image * temp = images_[original_position]->lazy_copy(W, H, FL_COLOR_EFFECT_NONE);
      images_[original_position]->reference();
      temp->add_(position);
      return temp;
    }
    // original might have its original too, we will rather refere to that than some intermediate
    Fl_Registered_Image * orig =  original_image_; // it might be itself, its OK
    Fl_Color_Effect new_lazy = fl_color_effect_multiply(lazy_effect_, additional_effect);
    while(orig!=orig->original_image_){
      new_lazy = fl_color_effect_multiply(orig->lazy_effect_, new_lazy);
      orig = orig->original_image_;
    }
    Fl_Registered_Image * temp = orig->lazy_copy(0,0, new_lazy);
    orig->reference();
    temp->add_(original_position);
    if(!W || (temp->w()==W && temp->h()==H))
      return temp;
    position++;
    temp = temp->lazy_copy(W, H, FL_COLOR_EFFECT_NONE);
    temp->add_(position);
    return temp;
}



void Fl_Registered_Image::draw(int X, int Y, int W, int H, int cx, int cy){
    update_image();
    if(!image_){
      Fl_Image::draw(X, Y, W, H, cx, cy);
      return;
    }
    if(original_image_)
        set_fast_original(image_, original_image_);
    image_->draw(X, Y, W, H, cx, cy);
    set_fast_original(0,0);
}

// testing purposes
void fl_display_shared_(){
  printf("\n Ind Orig O.Ind img D Flags   Effect Lazy-Ef.  Refc    W    H  Name  (total %d)\n", Fl_Registered_Image::num_images_);
  for(int i = 0; i<Fl_Registered_Image::num_images_; i++){
    Fl_Registered_Image *r = Fl_Registered_Image::images_[i];
    int cs = strlen(r->name_) - 15;
    if(cs<0) cs = 0;
    int original_index = -1;
    for(int j = 0; j<Fl_Registered_Image::num_images_; j++)
        if(r->original_image_ == Fl_Registered_Image::images_[j]){
        original_index = j;
           break;

        };
    const char * filled = r->image_? "o" : "x";
    const char * dt = (r->image_ && r->image_->data())? "o" : "x";
    printf("%4d%5d%6d  %s  %s%6X%9X%9X%6d%5d%5d  %s\n", i, r->original_, original_index, filled, dt,
           r->flags_, r->effect_, r->lazy_effect_, r->refcount_, r->w(), r->h(), r->name_ + cs);
  }
  printf("\n");
}

} // anonymous namespace

/*
Fl_Image * fl_original_image(Fl_Image * im){
    if(!fl_image_original_) return 0;
    fl_image_original_->update_image(); // we update it only on request, sometimes it might not have internal_image() yet
    return fl_image_original_->internal_image();
}
*/

// This is for testing porposes only - it prints table of all registered images and their properties
void fl_display_shared(){fl_display_shared_();} // work arrounf for namespace and friend


// This is only testing function, should be commented-out with release




///////////////// Fl_Shareable_Image  //////////////////////



Fl_Shared_Handler *Fl_Shareable_Image::handlers_ = 0;// Additional format handlers
int	Fl_Shareable_Image::num_handlers_ = 0;	// Number of format handlers
int	Fl_Shareable_Image::alloc_handlers_ = 0;	// Allocated format handlers

//void Fl_Shareable_Image::add();

int	Fl_Shareable_Image::refcount(){
    // For backward compatibility we should always return at least 1 to accomodate
    // for the shell so old applications know it can always call release() or delete
    // and destroy the shell - even if it does not have any internal pointer.
    return image_ ? ((Fl_Registered_Image *)image_)->refcount_ : 1;
}

void Fl_Shareable_Image::release(){
    if(image_ && ((Fl_Registered_Image *)image_)->to_remove==this){
        // ABI paranoya: we assure that this is not deleted if it was created
        //  by protected Fl_Shareable_Image(const char *n, Fl_Image *img = 0);
       ((Fl_Registered_Image *)image_)->release();
       return;
    }
    delete this;
}



void Fl_Shareable_Image::uncache(){
  if(image_)
    image_->uncache();

}


void Fl_Shareable_Image::update_from_registered(Fl_Image * r){
  image_ = r;
  update();

}

void Fl_Shareable_Image::init(){
  image_       = 0;
}


Fl_Shareable_Image::Fl_Shareable_Image() : Fl_Image(0,0,0) {
  init();
}


Fl_Shareable_Image::~Fl_Shareable_Image(){
  if(image_)
    ((Fl_Registered_Image *)image_)->release();
//  else
//    delete[] name_; // name_ was copied only if registered image does not exist
}





/** Returns the Fl_Shareable_Image* array */
Fl_Shareable_Image **Fl_Shareable_Image::images() {
  return (Fl_Shareable_Image **)Fl_Registered_Image::images_;
}

/** Returns the total number of shared images in the array. */
int Fl_Shareable_Image::num_images() {
  return Fl_Registered_Image::num_images_;
}


Fl_Shareable_Image::Fl_Shareable_Image(const char * name, int W, int H):Fl_Image(0, 0, 0){
    init();
    Fl_Registered_Image * i = Fl_Registered_Image::get(name, FL_COLOR_EFFECT_NONE, W, H);
    update_from_registered(i);
}

Fl_Shareable_Image::Fl_Shareable_Image(const Fl_Shareable_Image &src):Fl_Image(0, 0, 0){
  init();
  update_from_registered(RI(src.image_));
  if(image_) RI(image_)->reference();
}

static char name_buffer[64];
static int name_index_ = 0;
static const char * get_new_name(){
  sprintf(name_buffer, "/%d-fltk-internal-image", ++name_index_);
  return name_buffer;
}

Fl_Shareable_Image::Fl_Shareable_Image(Fl_Image * image, const char * name, int take_over):Fl_Image(0, 0, 0){
  init();
  if(image){
    if(fl_image_is_shared(image)){ // we handle this slightly differently otherwice shit happens...
      w(image->w());
      h(image->h());
      Fl_Registered_Image * ri = RI(((Fl_Shareable_Image *)image)->image_);
      if(ri){
        if(name){
          if(!ri->original_) ri = ri->original_image_;
          ri->reference();
          ri = ri->lazy_copy(0, 0, FL_COLOR_EFFECT_NONE, name);
          ri->effect_ = FL_COLOR_EFFECT_NONE;
          ri->add(0);
        }else
          ri->reference();
      }
      update_from_registered(ri);
      resize(image->w(), image->h());
      if(take_over) delete image;
      return;
    }
  }

  const char * un = name ? name : get_new_name();
  Fl_Registered_Image * temp = new Fl_Registered_Image(un, image);
  temp->add(0);
  temp->alloc_image_ = take_over;
  update_from_registered(temp);
  //if(!name)
  //  name_ = 0; // setting unique name to 0 for shell

}

Fl_Shareable_Image & Fl_Shareable_Image::operator = (const Fl_Shareable_Image & from){
    if (from.image_) ((Fl_Registered_Image *)(from.image_))->reference(); //we reference first
    if(image_) ((Fl_Registered_Image *)(from.image_))->release();
    update_from_registered(from.image_);
    return *this;
}



Fl_Shareable_Image * Fl_Shareable_Image::get(const char *n, int W, int H){
  Fl_Registered_Image * i = Fl_Registered_Image::get(n, FL_COLOR_EFFECT_NONE, W, H);
   if(!i) return 0;
   Fl_Shareable_Image * s = new Fl_Shareable_Image();
   s->update_from_registered(i);
   return s;
}



Fl_Shareable_Image * Fl_Shareable_Image::find(const char *n, int W, int H){
   Fl_Registered_Image * i = Fl_Registered_Image::find(n, FL_COLOR_EFFECT_NONE, W, H);
   if(!i) return 0;
   Fl_Shareable_Image * s = new Fl_Shareable_Image();
   s->update_from_registered(i);
   return s;
}



// This protected constructor only to assure backward-comparibility behaviour
Fl_Shareable_Image::Fl_Shareable_Image(const char *n, Fl_Image *img): Fl_Image(0,0,0) {
//  name_ = 0;
  if(img){
     Fl_Registered_Image * image =  new Fl_Registered_Image(n, img);
     update_from_registered(image);
  }else{
//    refcount_    = 1;
//    original_    = 0;
    image_       = 0;
//    alloc_image_ = 0;
//    name_ = new char[strlen(n) + 1]; // we assure that such aliens have copied name
//    strcpy((char *)name_, n);
  }
}

void Fl_Shareable_Image::add(){
    if(!image_) return;
    ((Fl_Registered_Image *)image_)->add(this);
}

//
// 'Fl_Shareable_Image::update()' - Update the dimensions of the shared images.
//


void Fl_Shareable_Image::update() {
  if (image_) {
    w(image_->w());
    h(image_->h());
    d(image_->d());
    Fl_Image::data(0, -1); // we indicate that it is shareable image by setting count to -1
  }
}

Fl_Image * Fl_Shareable_Image::copy(int W, int H) {
   Fl_Shareable_Image * n = new Fl_Shareable_Image(*this);
   // we have assure for different behaviour of copy() and resize().
   // When W = 0 :  - copy returns size of image it was copied from
   //               - resize() returns "full-scale" original
   if(W)
     n->resize(W, H);
   return n;
}

void Fl_Shareable_Image::resize(int W, int H){
    w(W);
    h(H);
    if(!image_) return;
    if(W==image_->w() && H == image_->h()) return; // nothing to do
    Fl_Registered_Image * im = RI(image_); // old, to be released
    Fl_Registered_Image * new_im = im->get(W, H, FL_COLOR_EFFECT_NONE);
    update_from_registered(new_im);
    im->release();
}

void Fl_Shareable_Image::assure_data() const {
    if(!image_) return;
    RI(image_)->update_image();
    // RI(image_)->flags_ |= Fl_Registered_Image::MUST_HAVE_DATA;
    //update();
}
void Fl_Shareable_Image::color_average(Fl_Color color, float mix){
    if(!image_) return;
    if(RI(image_)->is_bitmap()) return; // bitmaps dont color-average
    Fl_Color_Effect ea = fl_color_effect_average(color, mix);
    if(!ea) return;
    Fl_Registered_Image * im = RI(image_);
    Fl_Registered_Image * new_im = im->get(im->w(), im->h(), ea);
    update_from_registered(new_im);
    im->release();
}

void Fl_Shareable_Image::desaturate(){
    if(!image_) return;
    if(RI(image_)->is_bitmap()) return; // bitmaps dont desaturate
    Fl_Registered_Image * im = RI(image_);
    Fl_Registered_Image * new_im = im->get(im->w(), im->h(), FL_COLOR_EFFECT_DESATURATE);
    update_from_registered(new_im);
    im->release();
}

void Fl_Shareable_Image::draw(int X, int Y, int W, int H, int cx, int cy){
  if(image_)
    image_->draw(X, Y, W, H, cx, cy);
  else
    Fl_Image::draw(X, Y, W, H, cx, cy);
}


int Fl_Shareable_Image::original_w() const {
  return image_? RI(image_)->original_image_->w() : 0;
}

int Fl_Shareable_Image::original_h() const {
  return image_? RI(image_)->original_image_->h() : 0;
}


Fl_Image * Fl_Shareable_Image::original() const {
  if(Fl_Image::count()!= -1){ // not shareable image
    if(fl_original_image_ && fl_original_for_ == this){ // oh joy, we can find original fast!
      fl_original_image_->update_image(); // the data might not exist yet
      return fl_original_image_->image_;
    }
    // we need to look throught all shared images - slower
    for(int i = 0; i< Fl_Registered_Image::num_images_; i++){
      if(Fl_Registered_Image::images_[i]->image_ == this){
         if(Fl_Registered_Image::images_[i]->original_image_){
           Fl_Registered_Image::images_[i]->original_image_->update_image(); // the data might not exist yet
           return Fl_Registered_Image::images_[i]->original_image_->image_;
         }
         return 0;
      }
    }
    return 0;
  }
  // this is shareable image, we look for original
  if(image_ && RI(image_)->original_image_){
    RI(image_)->original_image_->update_image(); // the data might not exist yet
    return RI(image_)->original_image_->image_;
  }
  return 0;
}

const char * Fl_Shareable_Image::name() const {
  return image_ ? RI(image_)->name_ : 0;
}

///////////////////  unchanged  ////////////////

void Fl_Shareable_Image::add_handler(Fl_Shared_Handler f) {
  int			i;		// Looping var...
  Fl_Shared_Handler	*temp;		// New image handler array...

  // First see if we have already added the handler...
  for (i = 0; i < num_handlers_; i ++) {
    if (handlers_[i] == f) return;
  }

  if (num_handlers_ >= alloc_handlers_) {
    // Allocate more memory...
    temp = new Fl_Shared_Handler [alloc_handlers_ + 32];

    if (alloc_handlers_) {
      memcpy(temp, handlers_, alloc_handlers_ * sizeof(Fl_Shared_Handler));

      delete[] handlers_;
    }

    handlers_       = temp;
    alloc_handlers_ += 32;
  }

  handlers_[num_handlers_] = f;
  num_handlers_ ++;
}




void Fl_Shareable_Image::remove_handler(Fl_Shared_Handler f) {
  int	i;				// Looping var...

  // First see if the handler has been added...
  for (i = 0; i < num_handlers_; i ++) {
    if (handlers_[i] == f) break;
  }

  if (i >= num_handlers_) return;

  // OK, remove the handler from the array...
  num_handlers_ --;

  if (i < num_handlers_) {
    // Shift later handlers down 1...
    memmove(handlers_ + i, handlers_ + i + 1,
           (num_handlers_ - i) * sizeof(Fl_Shared_Handler ));
  }
}




void Fl_Shareable_Image::draw(Fl_Shareable_Image * im, int X, int Y, int W, int H){
  if(!im || !RI(im)->internal_image()) return;
  double fw = W;
  double fh = H;
  if(!W){
    if(H)
      fw = fh * double(im->original_w())/double(im->original_h());
    else{
      fw = im->w();
      fh = im->h();
    }
  }else if(!H)
    fh = fw * double(im->original_h())/double(im->original_w());
  W = int(fl_transform_dx(fw, 0) + 0.5);
  H = int(fl_transform_dy(0, fh) + 0.5);
  int old_w = im->w();
  int old_h = im->h();
  im->resize(W, H); // this is cheap if nothing happened
  // we need to push back original sizes
  RI(im)->set_sizes(old_w, old_h);
  im->draw(int(floor(fl_transform_x(X, Y))+0.5),int(floor(fl_transform_y(X, Y))+0.5), RI(im)->internal_image()->w(), RI(im)->internal_image()->h(), 0, 0);
}

void Fl_Shareable_Image::copy_function(Fl_Image * (*f)(Fl_Image * im, int W, int H)){
  copy_function_ = f;
}

void Fl_Shareable_Image::register_images(){Fl_Shareable_Image::add_handler(fl_check_images_);}














